package require Tk
package require treectrl

oo::class create MyOrdersView {
variable Names		; # допустимые колонки/ключи в данных
variable Defaults	; # значения "по умолчанию"
variable Grouping	; # имена в группировке
variable Columns    ; # имя => id_колонки
variable Rows		; # тикет=> id_строки
variable Orders		; # тикет=>dict ордер
variable Groups		; # группа=>id_строки
variable Window		; # 
variable ServerTime ;	
variable SystemTime ;
variable Timer
variable ParenTimer;  # проверяем жив-ли родитель (для тестера)
constructor { w } {
	set Names { ticket time type symbol price lots profit stopLoss takeProfit }
	set Grouping { symbol }
	array set Groups {}
	set Defaults {}
	foreach column $Names {
		dict set Defaults $column {}
	}
	array set Orders {} ; # изначально пусто
	set Window $w
	set ServerTime 0
	set SystemTime 0
	set Timer {}
	set ParenTimer {}
	my Layout
	my Timer
}

destructor {
	if { $Timer != {} } {
		catch { after cancel $Timer }
	}
	if { $Window != {} } {
		catch { $Window destroy }
	}
}
# вернуть дефолты ордера (может там время надо будет ставить, поэтому функция)
method Defaults {} {
	return $Defaults
}

# обновить данные про один ордер
# вернёт список обновлённых колонок
method Update { ticket keys args } {
	set r {}
	if { [ info exists Orders($ticket) ] } {
		set r $Orders($ticket)
		if { [ llength $args ] == {} } {
			return [ my DeleteOrder $ticket ]
		}
	} else {
		set r [ my Defaults ]
		if { [ llength $args ] == {} } {
			return {}
		}
	}
	set updates {}
	foreach { key value } $args {
		if { $key == "ticket" } {
			contunue
		}
		if { ! [ info exists Names($key) ] } {
			continue
		}
		if { [ dict get $Names($ticket) $key ] != $value } {
			dict set r $key $value
			lappend updates $key
		}
	}
	set Orders($ticket) $r
}

# создать/расположить внутренние виджеты
method Layout {} {
	set w $Window
	# создание и расположение основных виджетов
	toplevel $w
	wm withdraw $w
	wm title $w "Orders"
	wm protocol $w WM_DELETE_WINDOW [ list wm withdraw $w ]

	my Toolbar $w.tool
	
	treectrl $w.tree -yscrollcommand [ list $w.vscroll set ] -xscrollcommand [ list $w.hscroll set ]
	scrollbar $w.vscroll -orient vert -command [ list $w.tree yview ]
	scrollbar $w.hscroll -orient hor -command  [ list $w.tree xview ]
	my Statusbar $w.status
	grid $w.tool -row 0 -column 0 -columnspan 2 -sticky "ew"
	grid $w.tree -row 1 -column 0 -sticky "nsew"
	grid $w.vscroll -row 1 -column 1 -sticky "ns"
	grid $w.hscroll -row 2 -column 0 -sticky "ew"
	grid $w.status -row 3 -column 0 -columnspan 2 -sticky "ew"
	grid rowconfigure $w 1 -weight 1
	grid columnconfigure $w 0 -weight 1
	# настройка таблицы
	$w.tree configure -showheader 1 
	# элементы стилей в ячейках
	# текст и фон ;-)
	$w.tree element create elText text 
	$w.tree element create elBack rect
	# стиль для ячеек
	$w.tree style create styCell
	$w.tree style elements styCell {elText elBack}
	#	elBack
	$w.tree style layout styCell elBack -union elText
	foreach col $Names {
		set Columns($col) [ $w.tree column create -tags $col -itemstyle styCell ]
		$w.tree header configure 0 $Columns($col) -text $col
	}
	$w.tree configure -treecolumn $Columns(ticket)
	my Show
}
method Toolbar { w } {
	ttk::frame $w
	ttk::label $w.serverTime -textvariable [ my varname ServerTime ]
	pack {*}[winfo children $w] -side left
}
method Statusbar { w } {
	ttk::frame $w
	ttk::label $w.systemTime
	pack {*}[winfo children $w] -side left
}
method Show {} {
	wm deiconify $Window
	raise $Window
}
# получить группу из переданных параметров
# в соответсвии с Grouping
method Group { keys args } {
	set g {}
	foreach name $Grouping {
		set i [ lsearch -exact $keys $name ]
		if { $i < 0 } {
			return {}
		}
		lappend g [ lindex $args $i ]
	}
	if { [ llength $g ] != [ llength $Grouping ] } {
		return {}
	}
	return $g
}
# создать все узлы дерева для группы
# и вернуть нижний элемент
method TouchGroup { group } {
	if { [ llength $group ] == 0 } return {}
	set fqn {}
	set parent root
	foreach name $group {
		lappend fqn $name
		if { ! [ info exists Groups($fqn) ] } {
			set Groups($fqn) [ set item [ $Window.tree item create -button true -open true ] ]
			$Window.tree item text $item $Columns(ticket) $name
			$Window.tree item lastchild $parent $item
		}
		set parent $Groups($fqn)
	}
	return $Groups($fqn)
}
method Insert { keys args } {
	# вытащить из параметров ticket - он обязательный
	set i [ lsearch -exact $keys ticket ]
	if { $i < 0 } {
		return
	}
	set ticket [ lindex $args $i ]
	if { $ticket == {} } {
		return
	}
	#
	set Rows($ticket) [ set item [ $Window.tree item create ] ]
	foreach name $keys value $args {
		if { $name=="" || ! [ info exists Columns($name) ] } {
			continue
		}
		$Window.tree item text $item $name $value
	}
	set parent [ my TouchGroup [ my Group $keys {*}$args ] ]
	if { $parent == {} } {
		set parent root
	}
	$Window.tree item lastchild $parent $item
}

method Update { keys args } {
	set i [ lsearch -exact $keys ticket ]
	if { $i < 0 } {
		return
	}
	set ticket [ lindex $args $i ]
	if { $ticket == {} || ! [ info exists Rows($ticket) ] } {
		return
	}
	set item $Rows($ticket)
	foreach name $keys value $args {
		if { $name=="" || ! [ info exists Columns($name) ] } {
			continue
		}
		$Window.tree item text $item $name $value
	}
}

method Delete { keys args } {
	set i [ lsearch -exact $keys ticket ]
	if { $i < 0 } {
		return
	}
	set ticket [ lindex $args $i ]
	if { $ticket == {} } {
		return
	}
	if { $ticket == {} || ! [ info exists Rows($ticket) ] } {
		return
	}
	set item $Rows($ticket)
	$Window.tree item delete $item
	unset Rows($ticket);
}
method ServerTime { time } {
	set ServerTime $time
}
method Timer {} {
	set now [ clock seconds ]
	if { $now != $SystemTime } {
		set SystemTime $now
		$Window.status.systemTime configure -text [ clock format $SystemTime ]
	}
	set Timer [ after 30 [ list [ self ] Timer ] ]
}

# раз в 15 сек проверяем жив-ли родитель.
# тестер мог остановить процесс не вызывая OnDeinit
# переменную ::parent заведёт нить создатель и поместит туда свой id
method ParenTimer {} {
	if { [ info exists ::parent ] && $::parent!={} && ! [ thread exists $::parent ] } {
		thread::release [ thread::id ]	
	} else {
		set ParenTimer [ after 1500 [ list [ self ] ParenTimer ] ] 
	}	
}

export Insert Update Delete Show ServerTime 
export Timer ParenTimer

}; #/class MyOrdersView

wm withdraw .
set orders [ MyOrdersView create orders .orders ]
